﻿// Le Transit-O-Tron : comment ameliorer son transit en parapente
//
//===========================================================================================
// Copyright (c) 2011 Thierry Bernier
// A utiliser a tes risques et périls : ton disque dur pourrait bien prendre feu.
// Tu peux utiliser ce code source comme bon te semble, du moment que ce copyright
// est conserve.
//===========================================================================================


using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;


namespace Modele
{
    // Un point d'une trajectoire
    public class Point
    {
        public Point () { X = H = 0.0; }
        public Point (double iX, double iH) { X = iX; H = iH; }

        public double X, H;     // abscisse curviligne, hauteur                     (m)

        // Un peu d'algebre
        public static Point operator - (Point a, Point b) { return new Point (a.X - b.X, a.H - b.H); }
        public static double operator * (Point a, Point b) { return a.X * b.X + a.H * b.H; }
//      public Point Ortho () { return new Point (-H, X); }
    }

    // Vitesse sur trajectoire
    public class Speed
    {
        public Speed() { X = H = 0.0; }
        public Speed (double iX, double iH) { X = iX; H = iH; }

        public double X, H;     // suivant l'abscisse curviligne, et la hauteur     (m/s)

        public void Add (Speed s)
        {
            X += s.X;
            H += s.H;
        }
    }

    // Un segment de droite
    public class Segment
    {
        public Segment (Point iA, Point iB) { A = iA; B = iB; }
        public Point A, B;
    }

    // Un mobile a un instant : position, vitesse
    public class Mobile
    {
        public Point P;         // m
        public Speed V;         // m/s
        public double T;        // s

        // Avancer a la position de la date next (sans effet si dans le passe)
        public void Forward (double next)
        {
            if (next <= T) return;
            double d = next - T;
            P.X += V.X * d;
            P.H += V.H * d;
            T = next; 
        }

        public Mobile Copy ()
        {
            Mobile m = new Mobile ();
            m.P = new Point (P.X, P.H);
            m.V = new Speed (V.X, V.H);
            m.T = T;
            return m;
        }

    }

    // Trajectoire = une collection de points dates
    // l'abscisse curviligne peut revenir en arriere (vent fort de face)
    public class Trajet
    {
        public Trajet ()
        {
            mPositions = new List<Mobile>(100);
            mLastT = -999.0;
        }

        public void Append (Mobile pos)
        {
            if (pos.T > mLastT + 5.0)
            {
                mLastT = pos.T;
                mPositions.Add (pos);
            }
        }

        public List<Mobile> mPositions;
        private double mLastT;                   // debug
    }

    // Un pilote = une suite de regles de reaction aux vitesses et varios
    public class Pilote
    {
        // Pilote generique
        public Pilote ()
        {
            cmd = 0.0;
            known = null;
        }

        // Pilote a commande fixe
        public Pilote (double commande)
        {
            cmd = commande;
            known = null;
        }

        // Pilote fin (sur le taux de chute uniquement)
        public Pilote (Speed[] speed, double[] commande)
        {
            Debug.Assert (speed.Length == commande.Length);
            known = new List<KeyValuePair<double,double>>(speed.Length);
            for (int i = 0; i < speed.Length; ++i)
            {
                known.Add (new KeyValuePair<double,double>(speed[i].H, commande[i]));
            }
        }

        // Reaction de ce pilote a la vitesse sol actuelle
        public double Commande (Speed groundSpeed)
        {
            if (known == null)  return cmd;
            return Spline.Value (known, groundSpeed.H);
        }

        private double cmd;
        private List<KeyValuePair<double, double>> known;
    }

    // Une aile = une polaire des vitesses (pour chaque position d'accélérateur/trims ?)
    public class Aile
    {
        // Definie par 4 points caracteristiques de la polaire :
        // 0 = vitesse mini (juste avant decrochage)
        // 1 = taux de chute mini
        // 2 = vitesse bras-haut
        // 3 = accelerateur
        public Aile (Speed[] carac)
        {
            Caracteristique = new Speed[carac.Length];
            carac.CopyTo (Caracteristique, 0);
            Convert ();
        }

        public Aile ()
        {
            Caracteristique = new Speed[4];
            Caracteristique[0] = new Speed(24.0/3.6, -2.0);
            Caracteristique[1] = new Speed(31.0/3.6, -0.9);
            Caracteristique[2] = new Speed(39.0/3.6, -1.1);
            Caracteristique[3] = new Speed(58.0/3.6, -1.8);
            Convert ();
        }


        // Vitesse pour une commande donnee :
        //  -1.0 = frein a fond
        //  -0.3 = 30% de frein
        //   0.0 = bras haut
        //   0.5 = 1er barreau
        //   1.0 = accelerateur à fond
        // TODO : freiner + accelerer ...
        public Speed AirSpeed (double commande)
        {
            double k = (commande + 1.0) / 2.0;          // Revenir dans [0,1]
            double X =   Caracteristique[0].X * k
                       + Caracteristique[Caracteristique.Length-1].X * (1.0 - k);
            double H = Spline.Value (known, X);
            return new Speed (X, H);
        }

        public Speed[] Caracteristique { private set; get; }

        private List<KeyValuePair<double, double>> known;
        private void Convert ()
        {
            known = new List<KeyValuePair<double, double>> ();
            foreach (Speed s in Caracteristique)
                known.Add (new KeyValuePair<double,double>(s.X, s.H));
        }
    }

    // Une zone d'air bizarre
    // C'est un quadrilatere avec deux aretes horizontales (sommet et base)
    // La vitesse H,V du vent est definie au sommet et a la base (interpolee entre les deux)
    public class Zone
    {
        public Zone (double le, double ri, double lo, double hi)    // Un rectangle
        {
            Xhile = le;
            Xhiri = ri;
            Xlole = le;
            Xlori = ri;
            Hhi = hi;
            Hlo = lo;
            Vent = new Speed (0.0, 0.0);
            SetSegs();
        }

        public bool Inside (Point p)
        {
            return (p.X > Xlole) && (p.X < Xlori) && (p.H > Hlo) && (p.H < Hhi);
        }

        private void SetSegs()
        {
            SegLo = new Segment (new Point(Xlole, Hlo), new Point(Xlori, Hlo));
            SegHi = new Segment (new Point(Xhile, Hhi), new Point(Xhiri, Hhi));
            SegLe = new Segment (new Point(Xlole, Hlo), new Point(Xhile, Hhi));
            SegRi = new Segment (new Point(Xlori, Hlo), new Point(Xhiri, Hhi));
        }

        // Les quatres aretes de cette zone
        public Segment SegLo { get; private set; }
        public Segment SegHi { get; private set; }
        public Segment SegLe { get; private set; }
        public Segment SegRi { get; private set; }

        public double Xhile, Xhiri, Xlole, Xlori;  // Absicsses High, Low, Right, Left
        public double Hhi, Hlo;                      // Hauteur High, Low
        public Speed Vent;
    }

    // Un ciel = une collection de zones + du vent entre elles
    public class Sky
    {
        // Les limites du ciel (hors-jeu au dela)
        public static double clipMinX = -1.0;
        public static double clipMaxX = 10000.0;
        public static double clipMinH = -1.0;
        public static double clipMaxH = 3000.0;

        public Sky ()
        {
            mZones = new List<Zone> (10);
            mVent = new Speed ();
            Point lole = new Point (clipMinX, clipMinH);
            Point lori = new Point (clipMaxX, clipMinH);
            Point hile = new Point (clipMinX, clipMaxH);
            Point hiri = new Point (clipMaxX, clipMaxH);
            mSegLo = new Segment (lole, lori);
            mSegHi = new Segment (hile, hiri);
            mSegLe = new Segment (lole, hile);
            mSegRi = new Segment (lori, hiri);
        }

        // Calcul de la trajectoire en partant de l'altitude alt
        public Trajet Fly (double alt, Pilote pilote, Aile aile)
        {
            Trajet result = new Trajet ();

            // Point courant
            Mobile cur = new Mobile ();
            cur.P = new Point (0.0, alt);
            cur.V = new Speed (36.0 / 3.6, -2.00);
            result.Append (cur.Copy());     // 1er = depart

            // Tant que l'on est dans l'aire de jeu
            int guard = 300;
            while ((--guard > 0) && Inside (cur.P))
            {
                // Trouver le vent à la position courante
                Speed windspeed;
                Wind (cur.P, out windspeed);

                // Trouver la reaction du pilote face a ce vent, i.e. la vitesse air de l'aile
                double commande = pilote.Commande(windspeed);

                // Trouver la nouvelle vitesse air suite a la commande pilote (on n'itere pas ...)
                Speed vsair = aile.AirSpeed (commande);

                // Trouver la vitesse sol
                cur.V = vsair;
                cur.V.Add (windspeed);

                // Aller jusqu'a la prochaine zone avec cette vitesse sol
                MoveNext (cur);
                result.Append (cur.Copy());
            }

            return result;
        }

        public Zone ZoneContaining (Point p)
        {
            foreach (Zone z in mZones)
                if (z.Inside (p))
                    return z;
            return null;
        }


        private bool Inside (Point p)
        {
            return (p.X > clipMinX) && (p.X < clipMaxX) && (p.H > clipMinH) && (p.H < clipMaxH);
        }

        // Somme des vents des zones contenant le point donne
        private void Wind (Point p, out Speed v)
        {
            v = new Speed();

            // a) le vent de chaque zone contenant p
            bool seen = false;
            foreach (Zone z in mZones)
                if (z.Inside (p))
                {
                    seen = true;
                    v.Add (z.Vent);
                }

            // b) Le vent general si on est hors de tout thermique
            if (!seen) v.Add (mVent);
        }

        // Deplacer p a la plus proche frontiere sur la trajectoire courante
        private void MoveNext (Mobile p)
        {
            double t;
            double mint = p.T + 2.0e6;

            // a) recherche parmi les zones
            foreach (Zone z in mZones)
            {
                if (((t = ETA (p, z.SegLo)) > 0.0) && (t < mint)) mint = t;
                if (((t = ETA (p, z.SegHi)) > 0.0) && (t < mint)) mint = t;
                if (((t = ETA (p, z.SegRi)) > 0.0) && (t < mint)) mint = t;
                if (((t = ETA (p, z.SegLe)) > 0.0) && (t < mint)) mint = t;
            }

            // b) recherche dans le cadre
            if (((t = ETA (p, mSegLo)) > 0.0) && (t < mint)) mint = t;
            if (((t = ETA (p, mSegHi)) > 0.0) && (t < mint)) mint = t;
            if (((t = ETA (p, mSegRi)) > 0.0) && (t < mint)) mint = t;
            if (((t = ETA (p, mSegLe)) > 0.0) && (t < mint)) mint = t;

            // c) Nouvelle position
            if (mint < p.T+1.0e6) p.Forward (mint + 0.1);  // Un peu au dela de l'intersection trouvée
        }


        // Date a laquelle le mobile p coupe le segment s, rend <0.0 si pas d'intersection
        private double ETA (Mobile p, Segment s)
        {
            // L'intersection avec la droite portee par s doit etre dans s
            Point ap = p.P - s.A;
            Point hv = new Point (-p.V.H, p.V.X);
            Point hs = new Point (s.A.H-s.B.H, s.B.X-s.A.X);
            double l = (s.B - s.A) * hv;
            if (Math.Abs(l) < 1.0e-6) return -1.0;      // Vitesse parallele au segment
            l = (ap * hv) / l;
            if ((l < 0.0) || (l > 1.0)) return -1.0;    // Coupe hors du segment 
            l = (new Point (p.V.X, p.V.H)) * hs;
            if (Math.Abs (l) < 1.0e-6) return -1.0;     // Vitesse parallele au segment
            l = (ap * hs) / l;
            if (l >= 0.0) return -1.0;                  // Le segment est derriere le mobile
            return p.T-l;
        }

        public List<Zone> mZones;
        private Segment mSegLo, mSegHi, mSegLe, mSegRi;
        public Speed mVent;
    }


    // Copied from http://geekswithblogs.net/JuanDoNeblo/archive/2007/10/25/Data-Interpolation-with-SPLINE-in-Csharp.aspx
    public class Spline
    {
        public static double Value (List<KeyValuePair<double, double>> knownSamples, double z)
        {
            int np = knownSamples.Count;

            if (z <= knownSamples[0].Key) return knownSamples[0].Value;             // TODO BUG
            if (z >= knownSamples[np-1].Key) return knownSamples[np-1].Value;

            if (np > 1)
            {
                double[] a = new double[np];
                double x1;
                double x2;
                double y;
                double[] h = new double[np];

                for (int i = 1; i <= np - 1; i++)
                {
                    h[i] = knownSamples[i].Key - knownSamples[i - 1].Key;
                }

                if (np > 2)
                {
                    double[] sub = new double[np - 1];
                    double[] diag = new double[np - 1];
                    double[] sup = new double[np - 1];
                    for (int i = 1; i <= np - 2; i++)
                    {
                        diag[i] = (h[i] + h[i + 1]) / 3;
                        sup[i] = h[i + 1] / 6;
                        sub[i] = h[i] / 6;
                        a[i] = (knownSamples[i + 1].Value - knownSamples[i].Value) / h[i + 1] -
                               (knownSamples[i].Value - knownSamples[i - 1].Value) / h[i];
                    }

                    // SolveTridiag is a support function, see Marco Roello's original code
                    // for more information at
                    // http://www.codeproject.com/useritems/SplineInterpolation.asp
                    SolveTridiag (sub, diag, sup, ref a, np - 2);
                }

                int gap = 0;
                double previous = knownSamples[0].Key;

                // At the end of this iteration, "gap" will contain the index of the interval
                // between two known values, which contains the unknown z, and "previous" will
                // contain the biggest z value among the known samples, left of the unknown z
                for (int i = 0; i < knownSamples.Count; i++)
                {
                    if (knownSamples[i].Key < z && knownSamples[i].Key >= previous)
                    {
                        previous = knownSamples[i].Key;
                        gap = i + 1;
                    }
                }

                x1 = z - previous;
                x2 = h[gap] - x1;
                y = ((-a[gap - 1] / 6 * (x2 + h[gap]) * x1 + knownSamples[gap - 1].Value) * x2 +
                    (-a[gap] / 6 * (x1 + h[gap]) * x2 + knownSamples[gap].Value) * x1) / h[gap];

                return y;
            }

            return 0;
        }

        private static void SolveTridiag (double[] sub, double[] diag, double[] sup,
                                         ref double[] b, int n)
        {
            /*      solve linear system with tridiagonal n by n matrix a
                    using Gaussian elimination *without* pivoting
                    where   a(i,i-1) = sub[i]  for 2<=i<=n
                            a(i,i)   = diag[i] for 1<=i<=n
                            a(i,i+1) = sup[i]  for 1<=i<=n-1
                    (the values sub[1], sup[n] are ignored)
                    right hand side vector b[1:n] is overwritten with solution 
                    NOTE: 1...n is used in all arrays, 0 is unused */
            int i;
            /*                  factorization and forward substitution */
            for (i = 2; i <= n; i++)
            {
                sub[i] = sub[i] / diag[i - 1];
                diag[i] = diag[i] - sub[i] * sup[i - 1];
                b[i] = b[i] - sub[i] * b[i - 1];
            }
            b[n] = b[n] / diag[n];
            for (i = n - 1; i >= 1; i--)
            {
                b[i] = (b[i] - sup[i] * b[i + 1]) / diag[i];
            }
        }
    }
}

